/**
 * Detect the value whether is a function
 * @param {*} v the value to detect
 */
export function isFunction(v: any): boolean;
/**
 * Detect the value whether is a string
 * @param {*} v the value to detect
 */
export function isString(v: any): boolean;
/**
 * Detect the value whether is an object
 * @param {*} v the value to detect
 */
export function isObject(v: any): boolean;
/**
 * Detect the object whether is a pure object(the ctor is Object)
 * @param {*} v the value to detect
 */
export function isPureObject(v: any): boolean;
/**
 * get the parent class(ctor) of the ctor
 *
 * @callback
 * @param {Function} ctor
 * @returns {Function} the parent ctor
 */
export const getParentClass: typeof getSuperCtor;
/**
 * Abstract flat factory class
 * @abstract
 */
export class BaseFactory {
    /**
     * The Root Factory class
     * @name _Factory
     * @abstract
     * @internal
     * @type {typeof BaseFactory}
     */
    static _Factory: typeof BaseFactory;
    /**
     * The registered classes in the Factory
     * @name _children
     * @abstract
     * @internal
     * @type {{[name: string]:any|typeof BaseFactory}}
     */
    static _children: {
        [name: string]: any;
    };
    /**
     * the registered alias items object.
     * the key is alias name, the value is the registered name
     * @type {[alias: string]: string}
     * @abstract
     * @internal
     */
    static _aliases: [alias: string];
    /**
     * The Root Factory class
     */
    static get Factory(): typeof BaseFactory;
    /**
     * Extracts a specified number of words from a PascalCase class name to use as a base name for registration,
     * only if no `name` is specified. The parameter value indicates the maximum depth of the word extraction.
     *
     * In JavaScript, class names use `PascalCase` convention where each word starts with a capital letter.
     * The baseNameOnly parameter is a number that specifies which words to extract from the class name as the base name.
     * If the value is 1, it extracts the first word, 2 extracts the first two words, and 0 uses the entire class name.
     * The base name is used to register the class to the factory.
     *
     * @example
     *   such as "JsonTextCodec" if baseNameOnly is 1, the first word "Json" will be extracted from "JsonTextCodec" as
     *   the base name. If baseNameOnly is 2, the first two words "JsonText" will be extracted as the base name. If
     *   baseNameOnly is 0, the entire class name "JsonTextCodec" will be used as the base name.
     * @name _baseNameOnly
     * @type {number}
     * @default 1
     * @internal
     */
    static _baseNameOnly: number;
    /**
     * find the real root factory
     *
     * You can overwrite it to specify your root factory class
     * or set _Factory directly.
     * @abstract
     * @internal
     * @returns {typeof BaseFactory|undefined} the root factory class
     */
    static findRootFactory(): typeof BaseFactory | undefined;
    /**
     * find the real root factory
     * @internal
     * @param {typeof BaseFactory} aClass the abstract root factory class
     * @returns {typeof BaseFactory|undefined}
     */
    static _findRootFactory(aClass: typeof BaseFactory): typeof BaseFactory | undefined;
    static getRealName(name: any): any;
    /**
     * get the unique name in the factory from an alias name
     * @param {string} alias the alias name
     * @returns {string|undefined} the unique name in the factory
     */
    static getRealNameFromAlias(alias: string): string | undefined;
    /**
     * format(transform) the name to be registered.
     *
     * defaults to returning the name unchanged. By overloading this method, case-insensitive names can be achieved.
     * @abstract
     * @internal
     * @param {string} aName
     * @returns {string}
     */
    static formatName(aName: string): string;
    /**
     * Get the unique(registered) name in the factory
     * @param {string|Function} aClass
     * @returns {string} the unique name in the factory
     */
    static getNameFrom(aClass: string | Function): string;
    /**
     * format(transform) the name to be registered for the aClass
     * @param {*} aClass
     * @param {number} [aBaseNameOnly]
     * @returns {string} the name to register
     * @internal
     */
    static formatNameFromClass(aClass: any, aBaseNameOnly?: number): string;
    /**
     * register the aClass to the factory
     * @param {typeof BaseFactory} aClass the class to register the Factory
     * @param {IBaseFactoryOptions|any} [aOptions] the options for the class and the factory
     * @returns {boolean} return true if successful.
     */
    static register(...args: any[]): boolean;
    /**
     * register the aClass to the factory
     * @internal
     * @param {typeof BaseFactory} aClass the class to register the Factory
     * @param {IBaseFactoryOptions|any} [aOptions] the options for the class and the factory
     * @returns {boolean} return true if successful.
     */
    static _register(aClass: typeof BaseFactory, aOptions?: IBaseFactoryOptions | any): boolean;
    /**
     * Check the name, alias or itself whether registered.
     * @param {string|undefined} aName the class name
     * @returns {false|typeof BaseFactory} the registered class if registered, otherwise returns false
     */
    static registeredClass(aName: string | undefined): false | typeof BaseFactory;
    /**
     * unregister this class in the factory
     * @param {string|Function|undefined} aName the registered name or class, no name means unregister itself.
     * @returns {boolean} true means successful
     */
    static unregister(aName: string | Function | undefined): boolean;
    /**
     * remove all aliases of the registered item or itself
     * @param {string|typeof BaseFactory|undefined} aName the registered item or name
     */
    static cleanAliases(aName: string | typeof BaseFactory | undefined): void;
    /**
     * remove specified aliases
     * @param {...string} aliases the aliases to remove
     */
    static removeAlias(...aliases: string[]): void;
    /**
     * set aliases to a class
     * @param {typeof BaseFactory|string|undefined} aClass the class to set aliases
     * @param  {...string} aliases the left arguments are aliases
     *
     * @example
     *   import { BaseFactory } from 'custom-factory'
     *   class Factory extends BaseFactory {}
     *   const register = Factory.register.bind(Factory)
     *   const aliases = Factory.setAliases.bind(Factory)
     *   class MyFactory {}
     *   register(MyFactory)
     *   aliases(MyFactory, 'my', 'MY')
     *
     */
    static setAliases(aClass: typeof BaseFactory | string | undefined, ...aAliases: any[]): void;
    /**
     * set alias to a class
     * @param {typeof BaseFactory|string|undefined} aClass the class to set alias
     * @param {string} alias
     */
    static setAlias(aClass: typeof BaseFactory | string | undefined, alias: string): void;
    /**
     * get the aliases of the aClass
     * @param {typeof BaseFactory|string|undefined} aClass the class or name to get aliases, means itself if no aClass specified
     * @returns {string[]} aliases
     */
    static getAliases(aClass: typeof BaseFactory | string | undefined): string[];
    static set aliases(value: string[]);
    /**
     * the aliases of itself
     */
    static get aliases(): string[];
    /**
     * Get the display name from aClass
     * @param {string|Function|undefined} aClass the class, name or itself, means itself if no aClass
     * @returns {string|undefined}
     */
    static getDisplayName(aClass: string | Function | undefined): string | undefined;
    /**
     * Set the display name to the aClass
     * @param {string|Function|undefined} aClass the class, name or itself, means itself if no aClass
     * @param {string|{displayName: string}} aDisplayName the display name to set
     */
    static setDisplayName(aClass: string | Function | undefined, aDisplayName: string | {
        displayName: string;
    }): void;
    /**
     * This callback function is used to get all the classes in the factory. If 'brk' is returned, the function will stop.
     * @callback FactoryClassForEachFn
     * @param {typeof BaseFactory} ctor the class in the factory
     * @param {string} name the unique name in the factory
     * @returns {'brk'|string|undefined}
     */
    /**
     * executes a provided callback function once for each registered element.
     * @param {FactoryClassForEachFn} cb the forEach callback function
     * @returns {this}
     */
    static forEach(cb: (ctor: typeof BaseFactory, name: string) => 'brk' | string | undefined): any;
    /**
     * Get the registered class via name
     * @param {string} aName the class name or alias
     * @returns {undefined|typeof BaseFactory} return the registered class if found the name
     */
    static get(name: any): undefined | typeof BaseFactory;
    static _get(name: any): any;
    /**
     * Create a new object instance of Factory
     * @param {string|BaseFactory} aName
     * @param {*} aOptions
     * @returns {BaseFactory|undefined}
     */
    static createObject(aName: string | BaseFactory, aOptions: any): BaseFactory | undefined;
    constructor(...args: any[]);
    /**
     * initialize instance method
     * @abstract
     * @internal
     * @param {...*} [args] pass through all arguments coming from constructor
     */
    initialize(...args?: any[]): void;
}
export type IBaseFactoryOptions = {
    /**
     * the registered item name
     */
    name?: string | undefined;
    /**
     * the registered item display name
     */
    displayName?: string | undefined;
    baseNameOnly?: boolean | undefined;
    /**
     * the registered item aliases
     */
    aliases?: (string[] | string) | undefined;
    /**
     * the registered item alias
     */
    alias?: (string[] | string) | undefined;
    /**
     * defaults to true
     */
    isFactory?: Function | boolean;
    /**
     * defaults to true
     */
    autoInherits?: boolean;
};
import { getSuperCtor } from 'inherits-ex';
